# Python-SoXR
# High quality, one-dimensional sample-rate conversion library for Python.
# Python-SoXR is a Python wrapper of libsoxr.
# https://github.com/dofuuz/python-soxr

import warnings

import numpy as np

from . import cysoxr
from .cysoxr import QQ, LQ, MQ, HQ, VHQ
from ._version import version as __version__


__libsoxr_version__ = cysoxr.libsoxr_version()

# libsoxr locates memory per each channel.
# Too much channels will cause memory error.
_CH_LIMIT = 65536

_CONVERT_WARN_STR = 'Converting input to {}. Change ResampleStream/input dtype to avoid implicit conversion. This implicit conversion is deprecated and will be removed in next major update.'
_CH_EXEED_ERR_STR = 'Channel num({}) out of limit. Should be in [1, %d]' % _CH_LIMIT
_DTYPE_ERR_STR = 'Data type must be one of [float32, float64, int16, int32], not {}'
_QUALITY_ERR_STR = "Quality must be one of [QQ, LQ, MQ, HQ, VHQ]"


def _quality_to_enum(q):
    if q in (VHQ, HQ, MQ, LQ, QQ):
        return q

    if type(q) is int:
        raise ValueError(_QUALITY_ERR_STR)

    q = q.lower()
    if q in ('vhq', 'soxr_vhq'):
        return VHQ
    elif q in ('hq', 'soxr_hq'):
        return HQ
    elif q in ('mq', 'soxr_mq'):
        return MQ
    elif q in ('lq', 'soxr_lq'):
        return LQ
    elif q in ('qq', 'soxr_qq'):
        return QQ

    raise ValueError(_QUALITY_ERR_STR)


class ResampleStream:
    """ Streaming resampler

        Use `ResampleStream` for real-time processing or very long signal.

        Parameters
        ----------
        in_rate : float
            Input sample-rate.
        out_rate : float
            Output sample-rate.
        num_channels : int
            Number of channels.
        dtype : type or str, optional
            Internal data type processed with.
            Should be one of float32, float64, int16, int32.
        quality : int or str, optional
            Quality setting.
            One of `QQ`, `LQ`, `MQ`, `HQ`, `VHQ`.
    """

    def __init__(self,
                 in_rate: float, out_rate: float, num_channels: int,
                 dtype='float32', quality='HQ'):
        if in_rate <= 0 or out_rate <= 0:
            raise ValueError('Sample rate should be over 0')

        if num_channels < 1 or _CH_LIMIT < num_channels:
            raise ValueError(_CH_EXEED_ERR_STR.format(num_channels))

        self._type = np.dtype(dtype)
        if self._type not in (np.float32, np.float64, np.int16, np.int32):
            raise ValueError(_DTYPE_ERR_STR.format(self._type))

        q = _quality_to_enum(quality)

        self._cysoxr = cysoxr.CySoxr(in_rate, out_rate, num_channels, self._type.type, q)

    def resample_chunk(self, x, last=False):
        """ Resample chunk with streaming resampler

        Parameters
        ----------
        x : array_like
            Input array. Input can be 1D(mono) or 2D(frames, channels).
            If input is not `np.ndarray` or not dtype in constructor,
            it will be converted to `np.ndarray` with dtype setting.

        last : bool, optional
            Set True at end of input sequence.

        Returns
        -------
        np.ndarray
            Resampled data.
            Output is np.ndarray with same ndim with input.

        """
        if type(x) != np.ndarray or x.dtype != self._type:
            warnings.warn(_CONVERT_WARN_STR.format(self._type), DeprecationWarning, stacklevel=2)
            x = np.asarray(x, dtype=self._type)

        x = np.ascontiguousarray(x)    # make array C-contiguous

        if x.ndim == 1:
            y = self._cysoxr.process(x[:, np.newaxis], last)
            return np.squeeze(y, axis=1)
        elif x.ndim == 2:
            return self._cysoxr.process(x, last)
        else:
            raise ValueError('Input must be 1-D or 2-D array')


def resample(x, in_rate: float, out_rate: float, quality='HQ'):
    """ Resample signal

    Parameters
    ----------
    x : array_like
        Input array. Input can be 1D(mono) or 2D(frames, channels).
        If input is not `np.ndarray`, it will be converted to `np.ndarray(dtype='float32')`.
        Its dtype should be one of float32, float64, int16, int32.
    in_rate : float
        Input sample-rate.
    out_rate : float
        Output sample-rate.
    quality : int or str, optional
        Quality setting.
        One of `QQ`, `LQ`, `MQ`, `HQ`, `VHQ`.

    Returns
    -------
    np.ndarray
        Resampled data.
        Output is `np.ndarray` with same ndim and dtype with input.
    """
    if in_rate <= 0 or out_rate <= 0:
        raise ValueError('Sample rate should be over 0')

    if type(x) != np.ndarray:
        x = np.asarray(x, dtype=np.float32)

    if not x.dtype.type in (np.float32, np.float64, np.int16, np.int32):
        raise ValueError(_DTYPE_ERR_STR.format(x.dtype.type))

    q = _quality_to_enum(quality)

    x = np.ascontiguousarray(x)    # make array C-contiguous

    if x.ndim == 1:
        y = cysoxr.cysoxr_divide_proc(in_rate, out_rate, x[:, np.newaxis], q)
        return np.squeeze(y, axis=1)
    elif x.ndim == 2:
        num_channels = x.shape[1]
        if num_channels < 1 or _CH_LIMIT < num_channels:
            raise ValueError(_CH_EXEED_ERR_STR.format(num_channels))

        return cysoxr.cysoxr_divide_proc(in_rate, out_rate, x, q)
    else:
        raise ValueError('Input must be 1-D or 2-D array')


def _resample_oneshot(x, in_rate: float, out_rate: float, quality='HQ'):
    """
    Resample using libsoxr's `soxr_oneshot()`. Use `resample()` for general use.
    `soxr_oneshot()` becomes slow with long input.
    This function exists for test purpose.
    """
    return cysoxr.cysoxr_oneshot(in_rate, out_rate, x, _quality_to_enum(quality))




import numpy as np
import pytest
import soxr


@pytest.mark.xfail(raises=ValueError, strict=True)
@pytest.mark.parametrize('in_rate, out_rate', [(100, 0), (100, -1), (0, 100), (-1, 100)])
def test_bad_sr(in_rate, out_rate):
    x = np.zeros(100)
    soxr.resample(x, in_rate, out_rate)


@pytest.mark.parametrize('dtype', [np.float32, np.float64, np.int16, np.int32])
def test_dtype(dtype):
    x = np.random.randn(100).astype(dtype)

    y = soxr.resample(x, 100, 200)

    assert x.dtype == y.dtype


@pytest.mark.parametrize('in_rate, out_rate', [(44100, 32000), (32000, 44100)])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
def test_divide_match(in_rate, out_rate, dtype):
    x = np.random.randn(49999).astype(dtype)

    y_oneshot = soxr._resample_oneshot(x, in_rate, out_rate)
    y_divide = soxr.resample(x, in_rate, out_rate)

    assert np.allclose(y_oneshot, y_divide)


@pytest.mark.parametrize('in_rate, out_rate', [(44100, 32000), (32000, 44100)])
@pytest.mark.parametrize('length', [0, 1, 2, 99, 100, 101, 31999, 32000, 32001, 34828, 34829, 34830, 44099, 44100, 44101, 47999, 48000, 48001, 66149, 66150, 66151])
def test_length_match(in_rate, out_rate, length):
    x = np.random.randn(length).astype(np.float32)

    y_oneshot = soxr._resample_oneshot(x, in_rate, out_rate)
    y_divide = soxr.resample(x, in_rate, out_rate)

    assert np.allclose(y_oneshot, y_divide)


@pytest.mark.parametrize('channels', [1, 2, 3, 5, 7, 97, 197])
def test_channel_match(channels):
    x = np.random.randn(30011, channels).astype(np.float32)

    y_oneshot = soxr._resample_oneshot(x, 44100, 32000)
    y_divide = soxr.resample(x, 44100, 32000)

    assert np.allclose(y_oneshot, y_divide)


@pytest.mark.parametrize('in_rate, out_rate', [(44100, 32000), (32000, 44100)])
@pytest.mark.parametrize('dtype', [np.float32, np.float64])
@pytest.mark.parametrize('channels', [1, 2])
def test_stream_match(in_rate, out_rate, dtype, channels):
    CHUNK_SIZE = 509
    x = np.random.randn(49999, channels).astype(dtype)

    y_oneshot = soxr._resample_oneshot(x, in_rate, out_rate)

    rs_stream = soxr.ResampleStream(in_rate, out_rate, channels, dtype=dtype)

    y_list = []
    for idx in range(0, len(x), CHUNK_SIZE):
        end = idx + CHUNK_SIZE
        eof = False
        if len(x) <= end:
            eof = True
            end = len(x)
        y_chunk = rs_stream.resample_chunk(x[idx:end], last=eof)
        y_list.append(y_chunk)

    y_stream = np.concatenate(y_list)

    assert np.allclose(y_oneshot, y_stream)


def make_tone(freq, sr, duration):
    return np.sin(2 * np.pi * freq / sr * np.arange(int(sr * duration)))


@pytest.mark.parametrize('in_rate,out_rate', [(44100, 22050), (22050, 44100)])
@pytest.mark.parametrize('quality', ['VHQ', 'HQ', 'MQ', 'LQ', 'QQ'])
def test_quality_sine(in_rate, out_rate, quality):
    FREQ = 512.0
    DURATION = 2.0
    x = make_tone(FREQ, in_rate, DURATION)
    y = make_tone(FREQ, out_rate, DURATION)

    y_pred = soxr.resample(x, in_rate, out_rate, quality=quality)

    err = np.mean(np.abs(y-y_pred))
    assert err < 1e-5



"""
Python-SoXR setup.py for Cython build

Please refer to BUILDING.md for building instructions.
Do not run setup.py directly for the build process.
"""

import logging
import subprocess
import sys
import sysconfig

from distutils.ccompiler import get_default_compiler
from setuptools import setup, Extension
from setuptools.command.sdist import sdist
from wheel.bdist_wheel import bdist_wheel


SYS_LIBSOXR = False

# python -m build -C--build-option=--use-system-libsoxr
if '--use-system-libsoxr' in sys.argv:
    sys.argv.remove('--use-system-libsoxr')
    SYS_LIBSOXR = True


class CySoxrExtension(Extension):
    def __init__(self, *args, **kwargs):
        self._include = []
        super().__init__(*args, **kwargs)

    @property
    def include_dirs(self):
        import numpy
        return self._include + [numpy.get_include()]

    @include_dirs.setter
    def include_dirs(self, dirs):
        self._include = dirs


# Submodule versioning
def get_git_version(cwd=''):
    try:
        result = subprocess.run(
            ['git', 'describe', '--tags', '--always', '--dirty'],
            cwd=cwd, capture_output=True, check=True, text=True)

        ver = result.stdout.strip()
        return ver
    except Exception as e:
        logging.warning(f'Can not get {cwd} version: {e}')
    return None


CSOXR_VERSION_C = '''
#include "csoxr_version.h"
const char * libsoxr_version() { return "%s"; }
'''


def write_csoxr_version():
    ver = get_git_version('libsoxr')
    if ver:
        with open(f'src/soxr/_csoxr_version.c', 'wt') as f:
            f.write(CSOXR_VERSION_C % (ver))
        logging.info(f'libsoxr version: {ver}')


class CustomSDistCmd(sdist):
    def run(self):
        write_csoxr_version()
        super().run()


class CustomWheelCmd(bdist_wheel):
    def run(self):
        write_csoxr_version()
        super().run()


src_static = [
    'libsoxr/src/soxr.c',
    'libsoxr/src/data-io.c',
    'libsoxr/src/dbesi0.c',
    'libsoxr/src/filter.c',
    'libsoxr/src/fft4g64.c',
    'libsoxr/src/cr.c',

    # WITH_CR32
    'libsoxr/src/cr32.c',
    'libsoxr/src/fft4g32.c',

    # WITH_CR64
    'libsoxr/src/cr64.c',

    # WITH_VR32
    # 'libsoxr/src/vr32.c',

    # WITH_CR32S
    'libsoxr/src/cr32s.c',
    'libsoxr/src/pffft32s.c',
    'libsoxr/src/util32s.c',

    # WITH_CR64S
    # 'libsoxr/src/cr64s.c',
    # 'libsoxr/src/pffft64s.c',
    # 'libsoxr/src/util64s.c',

    # Cython wrapper
    'src/soxr/cysoxr.pyx',
    'src/soxr/_csoxr_version.c',  # csoxr libsoxr_version()
]

src_dynamic = [
    # Cython wrapper
    'src/soxr/cysoxr.pyx',
    'src/soxr/csoxr_version.c',  # libsoxr soxr_version()
]

compile_args = ['-DSOXR_LIB']

platform = sysconfig.get_platform()
if '-arm' in platform:
    compile_args.append('-mfpu=neon')
elif '-i686' in platform:
    compile_args.append('-msse')

if get_default_compiler() in ['unix', 'mingw32']:
    compile_args += ['-std=gnu99', '-Werror=implicit']

extensions = [
    CySoxrExtension(
        "soxr.cysoxr",
        src_static,
        include_dirs=['libsoxr/src', 'src/soxr'],
        language="c",
        extra_compile_args=compile_args)
]

extensions_dynamic = [
    CySoxrExtension('soxr.cysoxr', src_dynamic, language='c', libraries=['soxr'])
]


if __name__ == "__main__":
    from Cython.Build import cythonize

    if SYS_LIBSOXR:
        logging.info('Building Python-SoXR using system libsoxr...')
        setup(
            ext_modules=cythonize(extensions_dynamic, language_level='3'),
        )
    else:
        logging.info('Building Python-SoXR using bundled libsoxr...')
        setup(
            cmdclass={'sdist': CustomSDistCmd, 'bdist_wheel': CustomWheelCmd},
            ext_modules=cythonize(extensions, language_level='3'),
        )
